-- Hazara Khan's implementation of a "Need or Greed" loot distribution system
-- for pugs with a master looter using HazLoot

debug = true

HNG = {}

-- DATA STRUCTURES
local Present      = {} -- A sub-list of the master, of present raiders' mains
local InvertedP    = {} -- Inverted present list

-- this copy is a hack. Move it somewhere else!
local STATE  = {}
local STATES = {}
STATE.IDLE              = 0 -- Scanning, looting, whatever. Not bidding.
STATE.AUCTION_OPEN      = 1 -- Auction is are open
STATE.LASTCALL          = 2 -- Last call for bids
STATE.ROLL_OPEN         = 3 -- Bids are closed - no valid bids - gone to offspec rolls
STATE.BANKING           = 4 -- Item is bing banked
STATE.DUSTING           = 5 -- Item is bing dusted
STATE.CONFIRMED         = 6 -- Bid winner is being confirmed
STATE.AWARDING          = 7 -- Loot is being awarded
STATES = {"Need rolls open","Last call for need","Greed rolls open","Banking","Dusting","Confirming winner","Awarding loot!",}
STATES[STATE.IDLE] = "Idle"

-- -- -- STRATEGY INTERFACE -- -- -- --
function HNG:GetName()
  return "Need or Greed"
end

function HNG:Init()
end
function HNG:SaveState()
end

HNG.auction = "need"
HNG.roll    = "greed"
HNG.bank    = "vendor"
HNG.dust    = "disenchanting"

--- returns true if state change is valid, usage message otherwise
function HNG:IsValidStateChange(from, to)
  if from == to then return true
  elseif (to == STATE.AUCTION_OPEN  and from == STATE.IDLE) or
     (to == STATE.AUCTION_OPEN  and from == STATE.AUCTION_OPEN) or
     (to == STATE.AUCTION_OPEN  and from == STATE.LASTCALL) or
     (to == STATE.AUCTION_OPEN  and from == STATE.ROLL_OPEN) or
     (to == STATE.LASTCALL      and from == STATE.AUCTION_OPEN) or
     (to == STATE.ROLL_OPEN     and from == STATE.LASTCALL) or
     (to == STATE.CONFIRMED     and from == STATE.LASTCALL) or
     (to == STATE.DUSTING       and from == STATE.LASTCALL) or
     (to == STATE.BANKING       and from == STATE.LASTCALL) or
     (to == STATE.CONFIRMED     and from == STATE.ROLL_OPEN) or
     (to == STATE.DUSTING       and from == STATE.ROLL_OPEN) or
     (to == STATE.BANKING       and from == STATE.ROLL_OPEN) or
     (to == STATE.DUSTING       and from == STATE.BANKING) or
     (to == STATE.BANKING       and from == STATE.DUSTING) or
     (to == STATE.DUSTING       and from == STATE.IDLE) or
     (to == STATE.BANKING       and from == STATE.IDLE) or
     (to == STATE.CONFIRMED     and from == STATE.BANKING) or
     (to == STATE.CONFIRMED     and from == STATE.DUSTING) or
     (to == STATE.AUCTION_OPEN  and from == STATE.CONFIRMED) or
     (to == STATE.AWARDING      and from == STATE.CONFIRMED) or
     (to == STATE.LASTCALL      and from == STATE.CONFIRMED) or
     (to == STATE.BANKING       and from == STATE.CONFIRMED) or
     (to == STATE.DUSTING       and from == STATE.CONFIRMED) or
     (to == STATE.ROLL_OPEN     and from == STATE.CONFIRMED)
    then
      self:Debug("Valid going from: "..STATES[from].." to "..STATES[to])
      return true
  else
      self:Debug("NOT valid going from: "..STATES[from].." to "..STATES[to])
  end
  return HNG:GetNextStates(from)
end

function HNG:GetNextStates(state)
  local usage = "dunno"
  if state == STATE.IDLE then
    usage = "\n/hl |cff1eff00n|rext to open need rolls for the next item"
  elseif state == STATE.AUCTION_OPEN then
    usage = "\n/hl |cff1eff00l|rast|cff1eff00c|rall to issue a last call"
  elseif state == STATE.LASTCALL then
    usage =  "\n/hl |cff1eff00n|rext to re-open need rolls\n"
    usage = usage.."/hl |cff1eff00o|rff|cff1eff00s|rpec to go to greed rolls\n"
    usage = usage.."/hl |cff1eff00w|rinner <name1> to set the winner\n"
    usage = usage.."/hl |cff1eff00b|rank to send the item to the banker\n" --- oops
    usage = usage.."/hl |cff1eff00d|rust to send the item to the disenchanter"
  elseif state == STATE.BANKING then
    usage = " /hl |cffff1e00a|rward to bank "
  elseif state == STATE.DUSTING then
    usage = " /hl |cffff1e00a|rward to dust "
  elseif state == STATE.CONFIRMED then
    self:Print(" /hl |cff1eff00a|rward to distribute confirmed loot")
  end
  return usage
end

function HNG:GetHelp(state)
  local help = "I'm confused. Try reloading your UI with /rl"
  if state == STATE.IDLE then
    help = "click here to announce, or \'need\' to open bids"
  elseif state == STATE.AUCTION_OPEN then
    help = "click \'last call\' to close suicide bidding"
  elseif state == STATE.LASTCALL then
    help = "\'suicide\' to re-open, or go to \'off-spec\', \'bank\' or \'dust\'"
  elseif state == STATE.BANKING then
    help = "click \'confirm\' to send to the banker"
  elseif state == STATE.DUSTING then
    help = "click \'confirm\' to send to the duster"
  elseif state == STATE.CONFIRMED then
    help = "click \'confirm\' to send to the winner"
  end
  return help
end

function HNG:GetConfirmation(state)
  local confirmation = "Confirm"
  if state == STATE.LASTCALL then
    confirmation = "Suicide"
  elseif state == STATE.ROLL_OPEN then
    confirmation = "Off-spec Roll"
  elseif state == STATE.BANKING then
    confirmation = "Banker"
  elseif state == STATE.DUSTING then
    confirmation = "Disenchant"
  end
  return confirmation
end

--- Returns the first candidate in an array
function HNG:GetCandidate(array)
  if not array then return nil end
  if not Present then return nil end

  for i,candidate in ipairs(array) do
    array[i] = HNG:Capitalize(candidate)
    if HNG:IsPresent(candidate) then
      return candidate
    end
  end
  return nil
end

--- Add a new member to the (non-existent) master list.
function HNG:AddMember(member, role, position)
  return HNG:MarkAsPresent(member)
end

--- Is the toon a main or alt in this SK List?
-- @returns true if so, false if not, nil if bad input
function HNG:IsMember(member)
  return (InvertedP[member] and true or false)
end

--- Returns the raid role of a member (or alt)
-- @returns "N/A"
function HNG:GetRole(member)
  return "N/A"
end

--- Clears the present list
function HNG:Clear()
  Present      = {}
  InvertedP    = {}
end

function HNG:SetMasterList(members, alts)
  return 0
end

--- NOP
function HNG:RemoveMember(member, confirm)
  return false
end

--- Returns the main of a member
function HNG:GetMain(member)
  if not member then return nil end
  member = HNG:Capitalize(member)
  return member
end

function HNG:GetToonCount(member)
  return 1
end

function HNG:MarkAsPresent(member)
  if not member then return nil end
  member = HNG:Capitalize(member)
  if InvertedP[member] then return true end
  table.insert(Present,member)
  InvertedP[member] = #Present
  return #Present
end

--- Is the member present in the raid?
-- @param member The name of a main or alt
-- @returns true if the member is present, the name of an alt's main if member is
--          an alt or false if the member is not present
function HNG:IsPresent(member)
  if not member then return false end
  member = HNG:Capitalize(member)
  return (UnitInParty(member) or UnitInRaid(member))
end

function HNG:GetMemberAt(position)
  return Present[position]
end

--- Return the current position in the present list of a party or raid member
function HNG:GetMasterPosition(member)
  if nil == member then return nil end
  member = HNG:Capitalize(member)
  return InvertedP[member]
end

--- Suicide an active member, and adjust the master SK List.
--  @param member A main that has been marked as present, or one of their alts.
--  @return The new position in the list, nil if they were not found, or there's no
--          present list.
function HNG:Suicide(member)
  if not member then return nil end
  member = HNG:Capitalize(member)
  if not Present then return nil end
  return HNG:Drop(HNG:GetMain(member), Present)
end

--- "Drops" an active member to the bottom of the Present list (NOP)
function HNG:Drop(member)
  return InvertedP[member]
end

--- How many people are in the master list?
--  @return The size of the master list, 0 if empty.
function HNG:GetMasterListSize()
  return (#Present)
end

--- How many people are in the "present" list?
--  @return The size of the present list, 0 if empty.
function HNG:GetPresentCount()
  return (#Present)
end

--- Returns (cloned) arrays of the master and alts list, with their roles.
--  The lists returned can be persisted externally for subsequent restoration with
--   SetMasterList()
--  @return Cloned copies of the master SK table and the corresponding Alts with roles
--  @see SetMasterList
function HNG:GetMasterList()
  return HNG:GetPresentList(), {}
end

--- Return a (cloned) array of the list of members marked "present" for a raid.
--  The list returned is a copy. It should not therefore be manipulated with any
--  expectation of changes being reflected to the actual list, but is
--  appropriate for display, etc.
--  @return An array of {[position]="raider"} tables, in SK List order.
function HNG:GetPresentList()
  local list = {}
  if Present then 
    for i,main in pairs(Present) do
      table.insert(list,{["name"]    = main,
                         ["main"]    = main,
                         ["present"] = true,
                         ["master"]  = i,
                        })
    end
  end
  return list
end

--- Export the master list to some format
-- By default, returns a string of CSV, for each main: "Position","Main","Role","Alt1","Alt1Role","Alt2",..
-- TODO: write a corresponding import
-- @params format Currently ignored (defaults to CSV)
function HNG:ExportMasterList(format)
  return HNG:ExportPresentList(format)
end

--- Export the present list to some format
-- By default, returns a string of CSV, for each main: "Position","Main","Role"
-- @params format Currently ignored (defaults to CSV)
function HNG:ExportPresentList(format)
  local export = {}
  local output = ""
  if Present then
    for i,main in pairs(Present) do
      output = output.."\""..i.."\",\""..main.."\"\n"
    end
  end
  return output
end

--- NOP
function HNG:Bootstrap()
  return false
end

--- Player names have initial capitals, and all lowercase thereafter.
function HNG:Capitalize(unit)
  if nil == unit then return nil end
  if "string" ~= type(unit) then return unit end
  unit = string.upper(string.sub(unit,1,1))..string.lower(string.sub(unit,2))
  return unit
end

function HNG:Print(message)
  if HazLoot then
    HazLoot:Print(message)
  else
    print("[HNG] "..message)
  end
end

function HNG:Debug(message)
  if HazLoot then
    HazLoot:Debug(message)
  else
    print("[HNG][DEBUG] "..message)
  end
end
